import asyncio
import json
from asyncio import Future
from pathlib import Path
from typing import List, Type, TypeVar

from pydantic import BaseModel as PdModel

from web_foundation.app.service.service import Service
from web_foundation.kernel import NamedContext
from web_foundation.kernel.isolates.isolate import Isolate

AppConfig = TypeVar("AppConfig", bound=PdModel)


class App:
    config: AppConfig
    name: str
    isolates: List[Isolate]
    services: List[Service]
    ctx: NamedContext
    debug: bool

    def __init__(self, name: str, debug: bool = False):
        self.name = name
        self.ctx = NamedContext()
        self.services = []
        self.isolates = []
        self.debug = debug

    def add_isolate(self, isolate: Isolate):
        """
        Set isolate to app and set isolate debug
        :param isolate: isolate to apply
        :return: None
        """
        isolate.debug = self.debug
        self.isolates.append(isolate)

    def load_config(self, conf_path: Path, config_model: Type[AppConfig]) -> AppConfig:
        """
        Load app config to user in
        :param conf_path:
        :param config_model: BaseModel to cast json file to pydantc
        :return: None
        """
        with open(conf_path, "r") as _json_file:
            conf = config_model(**json.loads(_json_file.read()))
            self.config = conf
            self.ctx.app_config = conf

    def perform(self) -> List[Future]:
        """
        Call performs from isolates and return all isolates Futures
        :return: None
        """
        return [isolate.perform() for isolate in self.isolates]

    async def run(self):
        """
        Func to run app manually (without Runner)
        :return: None
        """
        return await asyncio.wait([isolate.perform() for isolate in self.isolates])

    def add_service(self, service: Service):
        """
        Set service to use in app
        Set service to named contex
        :param service: service to set
        :return: None
        """
        self.ctx.set_obj(service, kind=Service)
        self.services.append(service)
